@@ -20,5 +20,6 @@ def success(msg, debug=True):
         print(f"[SUCCESS]! {msg}")
 
 
-def lazy_debug(x, add=1):
-    return x % int(_math.log(x + 1) + add) == 0
+def lazy_debug(x, add=0):
+    _scale = int(_math.log(max(x, 1)) * _math.log(max(add, 1)))
+    return x % (_scale + 1) == 0
